/*
 * include/asm-microblaze/uaccess.h
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2006 Atmark Techno, Inc.
 */

#ifndef _ASM_UACCESS_H
#define _ASM_UACCESS_H

#include <linux/errno.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/mmu.h>


#define VERIFY_READ	0
#define VERIFY_WRITE	1


static inline int __check_vma(void *addr, unsigned long size)
{
	struct mm_struct *mm = current->mm;
	struct vm_area_struct * vma;
	
	if (!size) return 0;
	if (!mm) return 0; /* INIT_TASK has no mm */
	/* All kernel addresses are OK */
	if ((unsigned long)addr >= CONFIG_KERNEL_START) return 0;
	
	down_read(&mm->mmap_sem);
	vma = find_vma(mm, (unsigned int)addr);
	
	if (!vma) {
		up_read(&mm->mmap_sem);
		return -EFAULT;
	}

	if( ((unsigned int)addr + size) > (vma->vm_end)) {
		up_read(&mm->mmap_sem);
		return -EFAULT;
	}
	
	up_read(&mm->mmap_sem);
	return 0;
}

#define __range_ok(addr, size) _range_ok((unsigned long)(addr), (unsigned long)(size))

/* access_ok should check the range validity in common, not for the specific thread.*/
/* As for us any range is valid */
#define __access_ok(addr,size) (1)

#define access_ok(type,addr,size) __access_ok(addr,size)

extern inline int verify_area(int type, const void * addr, unsigned long size)
{
        return access_ok(type,addr,size)?0:-EFAULT;
}

extern inline int bad_user_access_length (void)
{
	return 0;
}

#define __get_user(__gu_var, __gu_ptr) \
({	  int __gu_ret = 0; \
	  volatile typeof(*(__gu_ptr)) *__gu_real_ptr = (__gu_ptr); \
	  if (sizeof(*(__gu_real_ptr)) > 8) __gu_ret = -EFAULT; else { \
	  if(__check_vma((void *)__gu_real_ptr, sizeof(*(__gu_real_ptr))) == 0) {__gu_var = *__gu_real_ptr;} \
		  else __gu_ret = -EFAULT;} \
	  __gu_ret; \
})


#define __get_user_bad()	(bad_user_access_length (), (-EFAULT))

#define __put_user(__gu_var, __gu_ptr) \
({	  int __gu_ret = 0; \
	  volatile typeof(*(__gu_ptr)) *__gu_real_ptr = (__gu_ptr); \
	  if (sizeof(*(__gu_real_ptr)) > 8) __gu_ret = -EFAULT; else {\
	  if(__check_vma((void *)__gu_real_ptr, sizeof(*(__gu_real_ptr))) == 0) {*__gu_real_ptr = __gu_var;}\
		else __gu_ret = -EFAULT; } \
	  __gu_ret; \
})

  
#define __put_user_bad()	(bad_user_access_length (), (-EFAULT))

#define put_user(x, ptr)	__put_user(x, ptr)
#define get_user(x, ptr)	__get_user(x, ptr)

static inline int memset_user(void *to, char val, unsigned int n);

static inline int memset_user(void *to, char val, unsigned int n) {
	volatile unsigned char *c_to = (unsigned char *)to;
	if (n) {
		while((volatile int)__put_user(val, c_to) == 0) {
			c_to++;
			n--;
			if(!n) break;
		}
	}
	return n;
}


#define __copy_to_user(to,from,n)		(copy_to_user(to,from,n))
#define __copy_from_user(to,from,n)		(copy_from_user(to,from,n))
#define __copy_to_user_inatomic(to,from,n)	(__copy_to_user(to,from,n))
#define __copy_from_user_inatomic(to,from,n)	(__copy_from_user(to,from,n))

#define copy_from_user(__cu_ptr_to, __cu_ptr_from, __cu_count) \
({ \
	volatile unsigned char *__cu_ptr = (unsigned char *)(__cu_ptr_from), *__cu_pvar = (unsigned char *)(__cu_ptr_to); \
	unsigned int ____cu_count = __cu_count; \
	if (____cu_count) { \
	    while(__get_user(*__cu_pvar, __cu_ptr) == 0) { \
		    __cu_pvar++; \
		    __cu_ptr++; \
		    ____cu_count--; \
		    if(!____cu_count) break; \
	    } \
	} \
	____cu_count; \
})

#define copy_to_user(__cu_ptr_to, __cu_ptr_from, __cu_count) \
({ \
	volatile unsigned char *__cu_ptr = (unsigned char *)(__cu_ptr_to), *__cu_pvar = (unsigned char *)(__cu_ptr_from); \
	unsigned int ____cu_count = __cu_count; \
	if (____cu_count) { \
    	    while(__put_user(*__cu_pvar, __cu_ptr) == 0) { \
		    __cu_pvar++; \
		    __cu_ptr++; \
		    ____cu_count--; \
		    if(!____cu_count) break; \
	    } \
	} \
	____cu_count; \
})


/*
 * The exception table consists of pairs of addresses: the first is the
 * address of an instruction that is allowed to fault, and the second is
 * the address at which the program should continue.  No registers are
 * modified, so it is entirely up to the continuation code to figure out
 * what to do.
 *
 * All the routines below use bits of fixup code that are out of line
 * with the main instruction path.  This means when everything is well,
 * we don't even have to jump over them.  Further, they do not intrude
 * on our cache or tlb entries.
 */
struct exception_table_entry
{
	unsigned long insn, fixup;
};

#include <asm/processor.h>
#include <linux/types.h>
#include <linux/kernel.h>

static inline unsigned long clear_user(void *addr, unsigned long size)
{
	if (access_ok(VERIFY_WRITE, addr, size)) {
		return memset_user(addr, 0, size);
	}
	if ((unsigned long)addr < TASK_SIZE) {
		unsigned long over = (unsigned long)addr + size - TASK_SIZE;
		return memset_user(addr, 0, size - over);
	}
	return size;
}

/* Returns 0 if exception not found and fixup otherwise.  */
extern unsigned long search_exception_table(unsigned long);
#if 1

	/* Must return:
	 *
	 * -EFAULT		for an exception
	 * count		if we hit the buffer limit
	 * bytes copied		if we hit a null byte
	 * (without the null byte)
	*/

static inline long strncpy_from_user(char *dst, const char *src, long count) {
	unsigned char *c_to = (unsigned char *)dst, *c_from = (unsigned char *)src;
	unsigned char val;
	long real_count = 0;
	long saved_count = count;
	int res;

	while(saved_count) {
		res = __get_user(val, c_from);
		if (res) return -EFAULT;
		*c_to = val;

		if(!val) return real_count;
		
		c_from++;
		c_to++;
		real_count++;		
		
		count--;
		if(!count) return saved_count;
	}
	
	return count;
}



/*
 * Return the size of a string (including the ending 0)
 *
 * Return 0 on exception, a value greater than count if too long
 */
static inline long strnlen_user(const char *src, long count) {
	unsigned char c_to, *c_from = (unsigned char *)src;
	unsigned int real_count = 0;
	long saved_count = count;
	int res;
		
	while(saved_count) {
		res = __get_user(c_to, c_from);
		if(res) return 0; /*bad address*/
		c_from++;
		count--;
		real_count++;
		if (!c_to) return real_count;
		if (!count) return saved_count+1; /* too long*/
	}
	return 0; /*should not reach here*/
}
#endif

#endif /* _ASM_UACCESS_H */

